--[[
Interface: 1.7.1.0 b10490

Copyright (C) GtX (Andy), 2021

Author: GtX | Andy
Date: 12.04.2021

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Not to be added to any mods / maps or modified from its current release form.
No changes are to be made to this script without permission from GtX | Andy

Darf nicht zu Mods / Maps hinzugefügt oder von der aktuellen Release-Form geändert werden.
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


SupplyTransportMissionManager = {}
SupplyTransportMissionManager.MAX_MISSIONS = 3 -- This can be increased after saving game for first time in the 'missions.xml' at 'missions.maxSupplyTransportMissions' if a user wishes.

local isDevelopmentVersion = false
local SupplyTransportMissionManager_mt = Class(SupplyTransportMissionManager)

local loadError
local supplyTransportMissionManager

local modName = g_currentModName
local modDirectory = g_currentModDirectory

local buildId = 1
local version = "0.0.0.0"

function SupplyTransportMissionManager:new(isServer, isClient)
    local self = setmetatable({}, SupplyTransportMissionManager_mt)

    self.isServer = isServer
    self.isClient = isClient

    self.buildId = buildId
    self.version = version

    self.modName = modName
    self.modDirectory = modDirectory

    self.missions = {}

    g_messageCenter:subscribe(MessageType.UNLOADING_STATIONS_CHANGED, self.onUnloadingStationsChanged, self)

    -- Testing and content creator commands for SP only. Use -cheats start parameter
    if self.isServer and (isDevelopmentVersion or g_addCheatCommands) then
        if g_registeredConsoleCommands["stGenerateMissions"] == nil then
            addConsoleCommand("stGenerateMissions", "Generate missions for testing only [clearExistingContracts]", "consoleCommandGenerateMissions", self)
        end

        if g_registeredConsoleCommands["stSetMaxMissions"] == nil then
            addConsoleCommand("stSetMaxMissions", "Set maximum number of missions that will be generated [maxMissions (Leave blank to reset)]", "consoleCommandSetMaxMissions", self)
        end
    end

    return self
end

function SupplyTransportMissionManager:unloadMapData()
    g_messageCenter:unsubscribeAll(self)

    if self.isServer and (isDevelopmentVersion or g_addCheatCommands) then
        if g_registeredConsoleCommands["stGenerateMissions"] ~= nil then
            removeConsoleCommand("stGenerateMissions")
        end

        if g_registeredConsoleCommands["stSetMaxMissions"] ~= nil then
            removeConsoleCommand("stSetMaxMissions")
        end
    end
end

function SupplyTransportMissionManager:generateMissions(dt)
    local maxMissions = SupplyTransportMissionManager.MAX_MISSIONS or 3

    if #self.missions < maxMissions then
        local missionType = g_missionManager:getMissionType("supplyTransport")
        local missionsGenerated = false

        if missionType ~= nil then
            local generationAttempts = 0

            repeat
                local fillTypeIndex = self:getRandomFilltype()

                generationAttempts = generationAttempts + 1

                if fillTypeIndex ~= nil then
                    local mission = missionType.class:new(self.isServer, self.isClient)

                    mission.type = missionType
                    mission.customEnvironment = modName -- Localise texts

                    if mission:init(fillTypeIndex, self:getRandomLitres(fillTypeIndex), self:getRandomNpcIndex()) then
                        g_missionManager:assignGenerationTime(mission)
                        mission:register()

                        table.insert(g_missionManager.missions, mission)
                        table.insert(self.missions, mission)

                        missionsGenerated = true
                        generationAttempts = 0

                        if isDevelopmentVersion then
                            local fillTypeTitle = mission:getFillTypeTitle()
                            local sellingStationTitle = mission:getSellingStationTitle()
                            local pricePerLitre = g_i18n:formatMoney(mission.pricePerLitre * 1000, 0, true, true)

                            print(string.format("  Debug: Delivery mission created - %s (%s):  %s", sellingStationTitle, fillTypeTitle, pricePerLitre))
                        end
                    else
                        mission:delete()
                    end
                end
            until (generationAttempts >= 6) or (#self.missions >= maxMissions)
        end

        if missionsGenerated then
            g_messageCenter:publish(MessageType.MISSION_GENERATED, {self})
        end
    end
end

function SupplyTransportMissionManager:updateMissions(dt)
    for _, mission in ipairs(self.missions) do
        if mission.getRemainingTime ~= nil and mission:getRemainingTime() <= 0 then
            if mission.status == AbstractMission.STATUS_RUNNING then
                mission:finish(false) -- Fail the mission
            elseif mission.status == AbstractMission.STATUS_STOPPED then
                mission:delete() -- Never accepted

                break
            end
        end
    end
end

function SupplyTransportMissionManager:addMissionToList(mission)
    ListUtil.addElementToList(self.missions, mission)
end

function SupplyTransportMissionManager:removeMissionFromList(mission)
    ListUtil.removeElementFromList(self.missions, mission)
end

function SupplyTransportMissionManager:onUnloadingStationsChanged()
    local unloadingStations = g_currentMission.storageSystem:getUnloadingStations()

    for i = #self.missions, 1, -1 do
        local mission = self.missions[i]

        if mission.sellingStation ~= nil and unloadingStations[mission.sellingStation] == nil then
            mission.sellpointDeleted = true -- Make sure no penalty is applied

            if mission.status == AbstractMission.STATUS_RUNNING then
                if isDevelopmentVersion then
                    print(string.format("  Debug: Selling Point '%s' for 'Delivery Mission' has been deleted, contract cancelled!", mission.sellingStation.stationName))
                end

                mission:finish(false) -- Fail the mission but no penalty for missing crops, just payout crops delivered
            else
                if isDevelopmentVersion then
                    print(string.format("  Debug: Selling Point '%s' for 'Delivery Mission' has been deleted, contract removed from list!", mission.sellingStation.stationName))
                end

                mission:delete() -- Delete the mission so it is removed from list
            end
        end
    end
end

function SupplyTransportMissionManager:getRandomFilltype()
    local numFillTypes = 0
    local validFillTypes = {}
    local fillTypesAdded = {}

    for _, unloadingStation in pairs(g_currentMission.storageSystem:getUnloadingStations()) do
        if unloadingStation.owningPlaceable ~= nil and unloadingStation.isSellingPoint then
            for fillTypeIndex, _ in pairs (unloadingStation.acceptedFillTypes) do
                if not fillTypesAdded[fillTypeIndex] then
                    fillTypesAdded[fillTypeIndex] = true

                    numFillTypes = numFillTypes + 1
                    validFillTypes[numFillTypes] = fillTypeIndex
                end
            end
        end
    end

    return validFillTypes[math.random(numFillTypes)]
end

function SupplyTransportMissionManager:getRandomLitres(fillTypeIndex)
    if fillTypeIndex ~= nil then
        if fillTypeIndex == FillType.COTTON then
            return math.random(1, 8) * 20000
        else
            local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex)

            -- If it could be delivered in bales then make them whole bales
            for _, baleType in ipairs (g_baleTypeManager.squareBales) do
                if baleType.fillTypeName == fillType.name then
                    return math.random(1, 100) * 4000
                end
            end
        end
    end

    -- If not a bale or cotton, make sure there is a mission that is 50000 litres or less
    for _, mission in ipairs (self.missions) do
        if mission.contractLiters <= 50000 then
            return math.random(5, 250) * 1000 -- 5000 > 250,000
        end
    end

    return math.random(5, 50) * 1000 -- 5000 > 50,000
end

function SupplyTransportMissionManager:getRandomNpcIndex()
    local validNPCS = {}
    local numNpcs = g_npcManager.numNpcs

    local minNpcCount = 19 + (SupplyTransportMissionManager.MAX_MISSIONS or 3) -- There is 19 default NPCS loaded with all maps
    local startIndex = (numNpcs >= minNpcCount) and minNpcCount or 1 -- Ignore 'maps_npcs.xml' (Default 19) if possible, some are titled incorrectly vs image

    for index = startIndex, numNpcs do
        local valid = true

        for _, mission in ipairs (self.missions) do
            if mission.npcIndex == index then
                valid = false
                break
            end
        end

        if valid then
            table.insert(validNPCS, index)
        end
    end

    return validNPCS[math.random(#validNPCS)] or math.random(numNpcs)
end

function SupplyTransportMissionManager:consoleCommandSetMaxMissions(maxMissions)
    if self.isServer then
        SupplyTransportMissionManager.MAX_MISSIONS = tonumber(MathUtil.clamp(maxMissions or 3, 1, 8))

        return string.format("Max Supply & Transport Contracts set to %d.", SupplyTransportMissionManager.MAX_MISSIONS)
    end

    return "'stSetMaxMissions' is a server only command"
end

function SupplyTransportMissionManager:consoleCommandGenerateMissions(clearExistingContracts)
    if self.isServer then
        local numRemoved = 0
        local numAvailable = 0
        local activeMissions = 0

        local clearMissions = Utils.stringToBoolean(clearExistingContracts)

        for i = #self.missions, 1, -1 do
            local mission = self.missions[i]

            if mission ~= nil then
                if mission.status == AbstractMission.STATUS_STOPPED then
                    if clearMissions then
                        mission:delete()
                        numRemoved = numRemoved + 1
                    else
                        numAvailable = numAvailable + 1
                    end
                else
                    activeMissions = activeMissions + 1
                end
            end
        end

        self:generateMissions(g_currentDt)

        return string.format("%d contracts have been cleared, %d new contracts created.", numRemoved, (#self.missions - numAvailable) - activeMissions)
    end

    return "'stGenerateMissions' is a server only command"
end

local function isManagerInitiated()
    return supplyTransportMissionManager ~= nil
end

local function inGameMenuContractsFrame_update(frame, dt)
    if frame.contractsList ~= nil and frame.contractsList.selectedIndex ~= nil then
        local contract = frame.contracts[frame.contractsList.selectedIndex]

        if contract ~= nil and contract.mission:isa(SupplyTransportMission) and not frame.supplyTransportMissionUpdating then
            if contract.mission.getRemainingTime ~= nil then
                frame.fieldBigText:setText(contract.mission:getRemainingTime(true))
            end
        end
    end
end

local function inGameMenuContractsFrame_updateDetailContents(frame, contractIndex)
    local contract = frame.contracts[contractIndex]

    if contract.finished and contract.mission:isa(SupplyTransportMission) then
        local mission = contract.mission
        local reimbursement = mission.reimbursement or 0

        if reimbursement > 0 then
            local total = reimbursement

            if mission.stealingCost ~= nil then
                total = total + (mission.stealingCost * -1)
            end

            frame.tallyBox:getDescendantByName("reimburse"):setText(g_i18n:formatMoney(reimbursement, 0, true, true))
            frame.tallyBox:getDescendantByName("total"):setText(g_i18n:formatMoney(total, 0, true, true))
        end
    end
end

local function inGameMenuContractsFrame_updateFieldContractInfo(frame, superFunc, mission)
    if not mission:isa(SupplyTransportMission) then
        superFunc(frame, mission)
    else
        frame.supplyTransportMissionUpdating = true

        local missionInfo = mission:getData()

        frame.titleText:setText(g_i18n:getText("fieldJob_contract") .. ": " .. missionInfo.jobType)
        frame.actionText:setText(missionInfo.action)
        frame.rewardText:setText(g_i18n:formatMoney(mission.reward, 0, true, true))
        frame.fieldBigText:setText(missionInfo.timeLimit or "")
        frame.contractDescriptionText:setText(missionInfo.description)
        frame.npcFieldBox:setVisible(false)
        frame.ownerOfText:setText("")
        frame.areaText:setText("")

        frame.supplyTransportMissionUpdating = false
    end
end

local function missionManager_loadMapData(missionManager, xmlFile)
    if isManagerInitiated() then
        -- Need to overwrite 'updateFieldContractInfo' here after 'Seasons' as it does not return the superFunc unless it is an 'AbstractFieldMission'. :-(

        InGameMenuContractsFrame.update = Utils.appendedFunction(InGameMenuContractsFrame.update, inGameMenuContractsFrame_update)
        InGameMenuContractsFrame.updateDetailContents = Utils.appendedFunction(InGameMenuContractsFrame.updateDetailContents, inGameMenuContractsFrame_updateDetailContents)
        InGameMenuContractsFrame.updateFieldContractInfo = Utils.overwrittenFunction(InGameMenuContractsFrame.updateFieldContractInfo, inGameMenuContractsFrame_updateFieldContractInfo)
    end
end

local function missionManager_unloadMapData(missionManager)
    missionManager.supplyTransportMissionManager = nil

    if isManagerInitiated() then
        supplyTransportMissionManager:unloadMapData()
        supplyTransportMissionManager = nil
    end
end

local function missionManager_loadFromXMLFile(missionManager, superFunc, xmlFilename)
    if not superFunc(missionManager, xmlFilename) then
        return false
    end

    if isManagerInitiated() then
        local xmlFile = loadXMLFile("missionsXML", xmlFilename)

        if xmlFile ~= nil then
            SupplyTransportMissionManager.MAX_MISSIONS = math.min(getXMLInt(xmlFile, "missions#maxSupplyTransportMissions") or 3, 8)

            delete(xmlFile)
        end

        for _, mission in ipairs (missionManager.missions) do
            if mission:isa(SupplyTransportMission) then
                supplyTransportMissionManager:addMissionToList(mission)
            end
        end
    end

    return true
end

local function missionManager_saveToXMLFile(missionManager, superFunc, xmlFilename)
    local success = superFunc(missionManager, xmlFilename)
    local xmlFile = loadXMLFile("missionsXML", xmlFilename)

    if xmlFile ~= nil then
        setXMLInt(xmlFile, "missions#maxSupplyTransportMissions", math.min(SupplyTransportMissionManager.MAX_MISSIONS or 3, 8))

        saveXMLFile(xmlFile)
        delete(xmlFile)
    end

    return success
end

local function missionManager_generateMissions(missionManager, dt)
    if isManagerInitiated() then
        supplyTransportMissionManager:generateMissions(dt)
    end
end

local function missionManager_updateMissions(missionManager, dt)
    if isManagerInitiated() then
        supplyTransportMissionManager:updateMissions(dt)
    end
end

local function missionManager_removeMissionFromList(missionManager, mission)
    if isManagerInitiated() then
        if mission ~= nil and mission.type.category == SupplyTransportMission.CATEGORY then
            supplyTransportMissionManager:removeMissionFromList(mission)
        end
    end
end

local function initSupplyTransportMissionManager()
    if g_missionManager ~= nil then
        if g_missionManager.supplyTransportMissionManager == nil then
            local mod = g_modManager:getModByName(modName)

            version = tostring(mod.version) or version

            if mod.modName == "FS19_SupplyTransportContracts" or mod.modName == "FS19_SupplyTransportContracts_update" and mod.author:len() == 3 then
                source(modDirectory .. "scripts/SupplyTranspostMission.lua")
                source(modDirectory .. "scripts/events/SupplyTranspostMissionFinishedEvent.lua")

                MissionManager.loadMapData = Utils.prependedFunction(MissionManager.loadMapData, missionManager_loadMapData)
                MissionManager.unloadMapData = Utils.prependedFunction(MissionManager.unloadMapData, missionManager_unloadMapData)
                MissionManager.loadFromXMLFile = Utils.overwrittenFunction(MissionManager.loadFromXMLFile, missionManager_loadFromXMLFile)
                MissionManager.saveToXMLFile = Utils.overwrittenFunction(MissionManager.saveToXMLFile, missionManager_saveToXMLFile)
                MissionManager.generateMissions = Utils.appendedFunction(MissionManager.generateMissions, missionManager_generateMissions)
                MissionManager.updateMissions = Utils.appendedFunction(MissionManager.updateMissions, missionManager_updateMissions)
                MissionManager.removeMissionFromList = Utils.appendedFunction(MissionManager.removeMissionFromList, missionManager_removeMissionFromList)

                supplyTransportMissionManager = SupplyTransportMissionManager:new(g_server ~= nil, g_client ~= nil)
                g_missionManager.supplyTransportMissionManager = supplyTransportMissionManager
            else
                loadError = {
                    startUpdateTime = 2000,

                    update = function(self, dt)
                        self.startUpdateTime = self.startUpdateTime - dt

                        if self.startUpdateTime < 0 then
                            local text = string.format(g_i18n:getText("ST_loadError", modName), modName, mod.author)

                            if g_dedicatedServerInfo == nil then
                                if not g_gui:getIsGuiVisible() then
                                    g_gui:showYesNoDialog({
                                        title = string.format("%s - Version %s", mod.title, version),
                                        text = text,
                                        dialogType = DialogElement.TYPE_LOADING,
                                        callback = self.openModHubLink,
                                        yesText = g_i18n:getText("button_ok"),
                                        noText = g_i18n:getText("button_modHubDownload")
                                    })
                                end
                            else
                                print("\n" .. text .. "\n    - https://farming-simulator.com/mods.php?lang=en&country=be&title=fs2019&filter=org&org_id=129652&page=0" .. "\n")
                                self.openModHubLink(true)
                            end
                        end
                    end,

                    openModHubLink = function(ignore)
                        if ignore == false then
                            local language = g_languageShort
                            local link = "mods.php?lang=en&country=be&title=fs2019&filter=org&org_id=129652&page=0"
                            if language == "de" or language == "fr" then
                                link = "mods.php?lang=" .. language .. "&country=be&title=fs2019&filter=org&org_id=129652&page=0"
                            end

                            openWebFile(link, "")
                        end

                        removeModEventListener(loadError)
                        loadError = nil
                    end
                }

                addModEventListener(loadError)
            end
        else
            print(string.format("  Error: Failed to load mod '%s', 'SupplyTransportMissionManager' was already loaded by '%s'!", modName, g_missionManager.supplyTransportMissionManager.modName or "N/A"))
        end
    else
        print("  Warning: MissionManager failed to load, unable to initiate 'SupplyTransportMissionManager'!")
    end
end

initSupplyTransportMissionManager()
